# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: patched_databricks_artifacts.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from mlflow.protos import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"patched_databricks_artifacts.proto\x12\x1bmlflow_databricks_artifacts\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"\x89\x02\n\x16\x41rtifactCredentialInfo\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x12\n\nsigned_uri\x18\x03 \x01(\t\x12O\n\x07headers\x18\x04 \x03(\x0b\x32>.mlflow_databricks_artifacts.ArtifactCredentialInfo.HttpHeader\x12\x41\n\x04type\x18\x05 \x01(\x0e\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialType\x1a)\n\nHttpHeader\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xaa\x02\n\x15GetCredentialsForRead\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x12\n\npage_token\x18\x03 \x01(\t\x1ax\n\x08Response\x12M\n\x10\x63redential_infos\x18\x02 \x03(\x0b\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialInfo\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\tJ\x04\x08\x01\x10\x02:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage\"\xab\x02\n\x16GetCredentialsForWrite\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x12\n\npage_token\x18\x03 \x01(\t\x1ax\n\x08Response\x12M\n\x10\x63redential_infos\x18\x02 \x03(\x0b\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialInfo\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\tJ\x04\x08\x01\x10\x02:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage*s\n\x16\x41rtifactCredentialType\x12\x11\n\rAZURE_SAS_URI\x10\x01\x12\x15\n\x11\x41WS_PRESIGNED_URL\x10\x02\x12\x12\n\x0eGCP_SIGNED_URL\x10\x03\x12\x1b\n\x17\x41ZURE_ADLS_GEN2_SAS_URI\x10\x04\x32\xb8\x03\n DatabricksMlflowArtifactsService\x12\xc6\x01\n\x15getCredentialsForRead\x12\x32.mlflow_databricks_artifacts.GetCredentialsForRead\x1a;.mlflow_databricks_artifacts.GetCredentialsForRead.Response\"<\xf2\x86\x19\x38\n4\n\x04POST\x12&/mlflow/artifacts/credentials-for-read\x1a\x04\x08\x02\x10\x00\x10\x03\x12\xca\x01\n\x16getCredentialsForWrite\x12\x33.mlflow_databricks_artifacts.GetCredentialsForWrite\x1a<.mlflow_databricks_artifacts.GetCredentialsForWrite.Response\"=\xf2\x86\x19\x39\n5\n\x04POST\x12\'/mlflow/artifacts/credentials-for-write\x1a\x04\x08\x02\x10\x00\x10\x03\x42,\n\x1f\x63om.databricks.api.proto.mlflow\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

_ARTIFACTCREDENTIALTYPE = DESCRIPTOR.enum_types_by_name['ArtifactCredentialType']
ArtifactCredentialType = enum_type_wrapper.EnumTypeWrapper(_ARTIFACTCREDENTIALTYPE)
AZURE_SAS_URI = 1
AWS_PRESIGNED_URL = 2
GCP_SIGNED_URL = 3
AZURE_ADLS_GEN2_SAS_URI = 4


_ARTIFACTCREDENTIALINFO = DESCRIPTOR.message_types_by_name['ArtifactCredentialInfo']
_ARTIFACTCREDENTIALINFO_HTTPHEADER = _ARTIFACTCREDENTIALINFO.nested_types_by_name['HttpHeader']
_GETCREDENTIALSFORREAD = DESCRIPTOR.message_types_by_name['GetCredentialsForRead']
_GETCREDENTIALSFORREAD_RESPONSE = _GETCREDENTIALSFORREAD.nested_types_by_name['Response']
_GETCREDENTIALSFORWRITE = DESCRIPTOR.message_types_by_name['GetCredentialsForWrite']
_GETCREDENTIALSFORWRITE_RESPONSE = _GETCREDENTIALSFORWRITE.nested_types_by_name['Response']
ArtifactCredentialInfo = _reflection.GeneratedProtocolMessageType('ArtifactCredentialInfo', (_message.Message,), {

  'HttpHeader' : _reflection.GeneratedProtocolMessageType('HttpHeader', (_message.Message,), {
    'DESCRIPTOR' : _ARTIFACTCREDENTIALINFO_HTTPHEADER,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.ArtifactCredentialInfo.HttpHeader)
    })
  ,
  'DESCRIPTOR' : _ARTIFACTCREDENTIALINFO,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.ArtifactCredentialInfo)
  })
_sym_db.RegisterMessage(ArtifactCredentialInfo)
_sym_db.RegisterMessage(ArtifactCredentialInfo.HttpHeader)

GetCredentialsForRead = _reflection.GeneratedProtocolMessageType('GetCredentialsForRead', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETCREDENTIALSFORREAD_RESPONSE,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForRead.Response)
    })
  ,
  'DESCRIPTOR' : _GETCREDENTIALSFORREAD,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForRead)
  })
_sym_db.RegisterMessage(GetCredentialsForRead)
_sym_db.RegisterMessage(GetCredentialsForRead.Response)

GetCredentialsForWrite = _reflection.GeneratedProtocolMessageType('GetCredentialsForWrite', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETCREDENTIALSFORWRITE_RESPONSE,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForWrite.Response)
    })
  ,
  'DESCRIPTOR' : _GETCREDENTIALSFORWRITE,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForWrite)
  })
_sym_db.RegisterMessage(GetCredentialsForWrite)
_sym_db.RegisterMessage(GetCredentialsForWrite.Response)

_DATABRICKSMLFLOWARTIFACTSSERVICE = DESCRIPTOR.services_by_name['DatabricksMlflowArtifactsService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.databricks.api.proto.mlflow\220\001\001\240\001\001\342?\002\020\001'
  _GETCREDENTIALSFORREAD.fields_by_name['run_id']._options = None
  _GETCREDENTIALSFORREAD.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _GETCREDENTIALSFORREAD._options = None
  _GETCREDENTIALSFORREAD._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _GETCREDENTIALSFORWRITE.fields_by_name['run_id']._options = None
  _GETCREDENTIALSFORWRITE.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _GETCREDENTIALSFORWRITE._options = None
  _GETCREDENTIALSFORWRITE._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForRead']._options = None
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForRead']._serialized_options = b'\362\206\0318\n4\n\004POST\022&/mlflow/artifacts/credentials-for-read\032\004\010\002\020\000\020\003'
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForWrite']._options = None
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForWrite']._serialized_options = b'\362\206\0319\n5\n\004POST\022\'/mlflow/artifacts/credentials-for-write\032\004\010\002\020\000\020\003'
  _ARTIFACTCREDENTIALTYPE._serialized_start=979
  _ARTIFACTCREDENTIALTYPE._serialized_end=1094
  _ARTIFACTCREDENTIALINFO._serialized_start=109
  _ARTIFACTCREDENTIALINFO._serialized_end=374
  _ARTIFACTCREDENTIALINFO_HTTPHEADER._serialized_start=333
  _ARTIFACTCREDENTIALINFO_HTTPHEADER._serialized_end=374
  _GETCREDENTIALSFORREAD._serialized_start=377
  _GETCREDENTIALSFORREAD._serialized_end=675
  _GETCREDENTIALSFORREAD_RESPONSE._serialized_start=458
  _GETCREDENTIALSFORREAD_RESPONSE._serialized_end=578
  _GETCREDENTIALSFORWRITE._serialized_start=678
  _GETCREDENTIALSFORWRITE._serialized_end=977
  _GETCREDENTIALSFORWRITE_RESPONSE._serialized_start=458
  _GETCREDENTIALSFORWRITE_RESPONSE._serialized_end=578
  _DATABRICKSMLFLOWARTIFACTSSERVICE._serialized_start=1097
  _DATABRICKSMLFLOWARTIFACTSSERVICE._serialized_end=1537
DatabricksMlflowArtifactsService = service_reflection.GeneratedServiceType('DatabricksMlflowArtifactsService', (_service.Service,), dict(
  DESCRIPTOR = _DATABRICKSMLFLOWARTIFACTSSERVICE,
  __module__ = 'patched_databricks_artifacts_pb2'
  ))

DatabricksMlflowArtifactsService_Stub = service_reflection.GeneratedServiceStubType('DatabricksMlflowArtifactsService_Stub', (DatabricksMlflowArtifactsService,), dict(
  DESCRIPTOR = _DATABRICKSMLFLOWARTIFACTSSERVICE,
  __module__ = 'patched_databricks_artifacts_pb2'
  ))


# @@protoc_insertion_point(module_scope)
